

import numpy as np
import statsmodels.api as sm

try:
    import piecewise_regression.r_squared_calc as r_squared_calc
    from piecewise_regression.main import Fit
except ImportError:
    import r_squared_calc
    from main import Fit


class ModelSelection:
    """
    Experimental - uses simple BIC based on simple linear model.
    """

    def __init__(
            self, xx, yy, max_breakpoints=10, n_boot=20,
            max_iterations=30, tolerance=10**-5,
            min_distance_between_breakpoints=0.01, min_distance_to_edge=0.02,
            verbose=True
            ):

        self.models = []

        self.stop = False

        if verbose:
            print("Running fit with n_breakpoint = 0 . . ")

        self.no_breakpoint_fit(xx, yy)

        for k in range(1, max_breakpoints+1):
            if verbose:
                print("Running fit with n_breakpoint = {} . . ".format(k))
            bootstrapped_fit = Fit(
                xx, yy, n_breakpoints=k, verbose=False,
                n_boot=n_boot, max_iterations=max_iterations,
                tolerance=tolerance,
                min_distance_between_breakpoints=min_distance_between_breakpoints,
                min_distance_to_edge=min_distance_to_edge)
            fit_summary = bootstrapped_fit.get_results()
            fit_summary["n_breakpoints"] = k
            self.models.append(fit_summary)

        self.summary()

    def summary(self):

        header = "\n{:^70}\n".format("Breakpoint Model Comparision Results")

        line_length = 100
        double_line = "=" * line_length + "\n"
        single_line = "-" * line_length + "\n"

        table_header_template = "{:<15} {:>12} {:>12} {:>12} \n"
        table_header = table_header_template.format(
            "n_breakpoints", "BIC", "converged", "RSS")
        table_row_template = "{:<15} {:>12.5} {:>12} {:>12.5} \n"

        table_contents = header
        table_contents += double_line

        table_contents += table_header
        table_contents += single_line

        for model in self.models:

            if model["converged"]:
                model_row = table_row_template.format(
                    model["n_breakpoints"], model["bic"],
                    str(model["converged"]), model["rss"])
            else:
                model_row = table_row_template.format(
                    model["n_breakpoints"], "", str(model["converged"]), "")

            table_contents += model_row

        print(table_contents)

        print("Min BIC (Bayesian Information Criterion) suggests best model")

    def no_breakpoint_fit(self, xx, yy):

        Z = np.array([xx])
        Z = Z.T
        Z = sm.add_constant(Z, has_constant='add')
        # Basic OLS fit
        results = sm.OLS(endog=np.array(yy), exog=Z).fit()

        # get the predicted values
        ff = [(results.params[0] + results.params[1] * x) for x in xx]

        # Get Rss
        rss, tss, r_2, adjusted_r_2 = r_squared_calc.get_r_squared(
            yy, ff, n_params=2)

        # Calcualte BIC
        n = len(xx)  # No. data points
        k = 2  # No. parameters
        bic = n * np.log(rss/n) + k * np.log(n)

        fit_data = {
            "bic": bic,
            "n_breakpoints": 0,
            "estimates": {},
            "converged": True,
            "rss": rss
        }

        fit_data["estimates"]["const"] = results.params[0]
        fit_data["estimates"]["alpha1"] = results.params[1]

        self.models.append(fit_data)


if __name__ == "__main__":
    pass



import numpy as np


def get_r_squared(yy, ff, n_params):

    n_data = len(yy)
    yy_mean = np.mean(yy)

    # Calculate residual and total sum of squares
    residual_sum_squares = 0
    total_sum_squares = 0
    for i in range(n_data):
        residual_sum_squares += (yy[i] - ff[i])**2
        total_sum_squares += (yy[i] - yy_mean)**2

    # R Squares
    r_squared = 1 - residual_sum_squares/total_sum_squares

    # Adjusted R squared
    adjusted_r_squared = 1 - (1-r_squared) * \
        (n_data-1) / (n_data - n_params - 1)

    return residual_sum_squares, total_sum_squares, r_squared, adjusted_r_squared




import numpy as np
import scipy
import math


def get_test_statistic(xx_davies, yy_davies, theta):
    """
    Compute a test statistic for the Davies test for the p-value of existence
    of a breakpoint. Based on Davies(1987) "Hypothesis Testing when a nuisance
    parameter is present only under the alternative".
    All the variables in this function are as named and described in that
    paper.

    :param xx_davies: Data series in x-axis (same axis as the breakpoints).
    :type xx_davies: list of floats

    :param yy_davies: Data series in y-axis.
    :type yy_davies: list of floats

    :param theta: A test value from within the range of data in xx
    :type theta: float

    """
    n = len(xx_davies)
    s_0 = 0
    s_1 = 0
    s_2 = 0
    s_3 = 0
    s_4 = 0

    for x in xx_davies:
        s_0 += x**2

        if x > theta:
            s_1 += x*(x-theta)
            s_3 += x - theta

        elif x < theta:
            s_2 += x*(x-theta)
            s_4 += x - theta

    a_hat = np.sum(yy_davies)/n
    b_hat = np.sum(xx_davies*yy_davies)/s_0
    V = s_1*s_2/s_0 + s_3*s_4/n
    S = 0
    for i in range(n):
        if xx_davies[i] > theta:
            S += (yy_davies[i] - a_hat - b_hat*xx_davies[i]) * \
                (xx_davies[i] - theta)
    S = S/(np.sqrt(np.abs(V)))
    return S


def davies_test(xx, yy, k=10, alternative="two_sided"):
    """
    Significance test for the existence of a breakpoint
    Null hypothesis is that there is no breakpoint, or that the change in
    gradient is zero.
    Alternative hypothesis is that there is a breakpoint, with a non-zero
    change in gradient.
    The change is gradient is a function of the breakpoint position.
    The breakpoint posiition is a nuisannce parameter that only exists in the
    alternative hypothesis.
    Based on Davies (1987), "Hypothesis Testing when a nuisance parameter is
    present only under the alternative".

    :param xx: Data series in x-axis (same axis as the breakpoints).
    :type xx: list of floats

    :param yy: Data series in y-axis.
    :type yy: list of floats

    :param k: A control parameter that determines the number of points to
        consider within the xx range.
    :type k: int

    :param alternative: Whether to consider a two-sided hypothesis test,
        or a one sided test with change of gradient greater or less than zero.
        For existence of a breakpoint, use "two-sided".
    :type alternative: str. One of "two_sided", "less", "greater"

    """
    # Centre the x values - makes no difference to existence of a breakpoint
    # The Davies test has this as an assumption
    xx_davies = xx - np.mean(xx)
    yy_davies = yy

    # As in Muggeo's R package "segmented", cut from second to second to last
    # data point
    # Need more data in the xx than in [L,U] for the test to work
    L = xx_davies[1]
    U = xx_davies[-2]

    # More thetas is better
    thetas = np.linspace(L, U, k)
    # For each value of theta, compute a test statistic
    test_stats = []
    for theta in thetas:
        test_stat = get_test_statistic(xx_davies, yy_davies, theta)
        test_stats.append(test_stat)
    if alternative == "two_sided":
        # Two sided test, M as defined by Davies
        M = np.max(np.abs(test_stats))
    elif alternative == "less":
        M = np.abs(np.min(test_stats))
    elif alternative == "greater":
        M = np.max(test_stats)

    # Use formulas from Davies
    V = 0
    for i in range(len(thetas)-1):
        V += np.abs(test_stats[i+1]-test_stats[i])

    p = scipy.stats.norm.cdf(-M) + V * np.exp(-.5*M **
                                              2) * 1/(np.sqrt(8*math.pi))

    if alternative == "two_sided":
        return p*2
    else:
        return p




import numpy as np


def validate_boolean(var, var_name):
    if isinstance(var, bool):
        return var
    else:
        raise ValueError(
            "{} must be a Boolean: True or False".format(var_name))


def validate_positive_integer(var, var_name):
    if isinstance(var, bool):
        raise ValueError("{} must be a positive Integer".format(var_name))
    if isinstance(var, int) and var > 0:
        return var
    else:
        raise ValueError("{} must be a positive Integer".format(var_name))


def validate_non_negative_integer(var, var_name):

    if isinstance(var, bool):
        raise ValueError("{} must be a non-negative Integer".format(var_name))

    if isinstance(var, int) and var >= 0:
        return var
    else:
        raise ValueError("{} must be a non-negative Integer".format(var_name))


def validate_positive_number(var, var_name):
    if isinstance(var, bool):
        raise ValueError("{} must be a Float".format(var_name))
    if (isinstance(var, float) or isinstance(var, int)) and var > 0:
        return var
    else:
        raise ValueError("{} must be a Float".format(var_name))


def validate_list_of_numbers(var, var_name, min_length):
    """
    Allowed types:
            List of integers of floats
            Numpy array of integers or floats
    """
    value_error_text = "{} must be a list of numbers with minimum length {}"
    value_error_text = value_error_text.format(var_name, min_length)

    # If its a list, convert it to a numpy array
    if isinstance(var, list):
        var = np.array(var)

    # If its not a numpy array at this point, raise a value error
    if not isinstance(var, np.ndarray):
        raise ValueError(value_error_text)

    # Check the array has numebrs in it
    if not np.issubdtype(var.dtype, np.number):
        raise ValueError(value_error_text)

    if len(var) < min_length:
        raise ValueError(value_error_text)

    return var



from piecewise_regression.data_validation import (
        validate_positive_number,
        validate_boolean,
        validate_positive_integer,
        validate_list_of_numbers,
        validate_non_negative_integer
    )
import unittest

import os
import sys
sys.path.insert(1, os.path.join(sys.path[0], '..'))


class TestDataValidation(unittest.TestCase):

    def test_positive_number(self):

        valid_inputs = [0.111, 3, 6770, 23.1, 7, 0.00001]

        invalid_inputs = [-1, 0, "hi", [1, 1], True, False, None]

        for valid in valid_inputs:
            var_return = validate_positive_number(valid, "test")
            self.assertEqual(var_return, valid)

        for invalid in invalid_inputs:
            test_kwargs = {"var": invalid, "var_name": "test"}
            self.assertRaises(
                ValueError,
                validate_positive_number,
                **test_kwargs)

    def test_boolean(self):

        valid_inputs = [True, False]

        invalid_inputs = [-1, 0, "hi", [1, 1], 22, 1.11]

        for valid in valid_inputs:
            var_return = validate_boolean(valid, "test")
            self.assertEqual(var_return, valid)

        for invalid in invalid_inputs:
            test_kwargs = {"var": invalid, "var_name": "test"}
            self.assertRaises(
                ValueError,
                validate_boolean,
                **test_kwargs)

    def test_positive_integer(self):

        valid_inputs = [1, 2, 31, 6543]

        invalid_inputs = [-1, 0, "hi", [1, 1], 1.11, True, False, None]

        for valid in valid_inputs:
            var_return = validate_positive_integer(valid, "test")
            self.assertEqual(var_return, valid)

        for invalid in invalid_inputs:
            test_kwargs = {"var": invalid, "var_name": "test"}
            self.assertRaises(
                ValueError,
                validate_positive_integer,
                **test_kwargs)

    def test_non_negative_integer(self):

        valid_inputs = [1, 2, 0, 31, 6543]

        invalid_inputs = [-1, "hi", [1, 1], 1.11, True, False, None]

        for valid in valid_inputs:
            var_return = validate_non_negative_integer(valid, "test")
            self.assertEqual(var_return, valid)

        for invalid in invalid_inputs:
            test_kwargs = {"var": invalid, "var_name": "test"}
            self.assertRaises(
                ValueError,
                validate_non_negative_integer,
                **test_kwargs)

    def test_list_of_numbers(self):

        valid_inputs = [
            [1, 1, 1],
            [0.1, 43, 12],
            [2, 1]
            ]

        invalid_inputs = [
            -1, "hi", [1, "h"], 1.11, True, False, None,
            [], [1]]

        for valid in valid_inputs:
            var_return = validate_list_of_numbers(valid, "test", 2)
            self.assertListEqual(list(var_return), list(valid))

        for invalid in invalid_inputs:
            test_kwargs = {
                "var": invalid, "var_name": "test",
                "min_length": 2
                }
            self.assertRaises(
                ValueError,
                validate_list_of_numbers,
                **test_kwargs)


if __name__ == '__main__':
    unittest.main()



import piecewise_regression.r_squared_calc as r_squared_calc
import numpy as np
import unittest
from importlib.machinery import SourceFileLoader

import os
import sys
sys.path.insert(1, os.path.join(sys.path[0], '..'))

DATA_SOURCE = "tests/data/data.txt"


class TestRSquared(unittest.TestCase):

    def test_some_data(self):

        ff = np.linspace(0, 10)
        yy = np.linspace(0, 10)

        rss, tss, r_2, adjusted_r_2 = r_squared_calc.get_r_squared(yy, ff, 1)

        self.assertEqual(r_2, 1)
        self.assertEqual(adjusted_r_2, 1)

        ff = [0.1, 0.8, 0.4, -4, 6, 12, 14, 1]
        yy = [1, 2, 3, 4, 5, 6, 6, 6]

        rss, tss, r_2, adjusted_r_2 = r_squared_calc.get_r_squared(yy, ff, 1)

        # Value calculated from sklearn's r2_score function
        r_2_from_sklearn = -6.405023255813953
        self.assertEqual(r_2_from_sklearn, r_2)

        data = SourceFileLoader('data', DATA_SOURCE).load_module()

        ff = np.array(data.BP_1_FF)
        yy = np.array(data.BP_1_YY)

        rss, tss, r_2, adjusted_r_2 = r_squared_calc.get_r_squared(yy, ff, 6)
        # Value calculated from sklearn's r2_score function
        r_2_from_sklearn = 0.9990626123719015
        self.assertEqual(r_2_from_sklearn, r_2)


if __name__ == '__main__':
    unittest.main()



from piecewise_regression.main import Fit
from piecewise_regression.model_selection import ModelSelection

import numpy as np
import unittest
from importlib.machinery import SourceFileLoader

import os
import sys
sys.path.insert(1, os.path.join(sys.path[0], '..'))

DATA_SOURCE = "tests/data/data.txt"


class TestFit(unittest.TestCase):

    def test_it_works_with_differnet_options(self):
        """
        Chekc the the n_breakpoints converged True/False boolean gives the
        same answer with model selection and fit
        """
        # This is influenced by random seeds, so might not pass
        # with different random seeds (although it doensnt mean it isn't
        # working)
        np.random.seed(2)

        data = SourceFileLoader('data', DATA_SOURCE).load_module()

        xx = np.array(data.MUGGEO_1_XX)
        yy = np.array(data.MUGGEO_1_YY)

        ms = ModelSelection(xx, yy)

        # For each n_breakpoints, chekc the ModelSelection vs fit results
        for n_breakpoints in range(1,10):
            fit = Fit(xx, yy, n_breakpoints=n_breakpoints, verbose=False)

            fit_converged = fit.get_results()["converged"]
            ms_converged = ms.models[n_breakpoints]["converged"]

            self.assertEqual(fit_converged, ms_converged)

if __name__ == '__main__':
    unittest.main()



from piecewise_regression.main import NextBreakpoints

import numpy as np
import unittest
from importlib.machinery import SourceFileLoader

import os
import sys
sys.path.insert(1, os.path.join(sys.path[0], '..'))

DATA_SOURCE = "tests/data/data.txt"


class TestNextBreakpoint(unittest.TestCase):

    def test_against_muggeo_r_package_data_1(self):
        """
        Muggeo uses slightly different packages and methods etc, so just check
        values are very close, not exact
        Starting from Muggeo's converged breakpoint values, I am iterating
        once, slight change
        The NextBreakpoint class is very vanilla, so in this example is getting
        in some local minima
        """

        data = SourceFileLoader('data', DATA_SOURCE).load_module()

        xx = np.array(data.MUGGEO_1_XX)
        yy = np.array(data.MUGGEO_1_YY)

        # Choose some bps values from Muggeo converged values
        bps = np.array([7.152, 12.161])

        next_fit = NextBreakpoints(xx, yy, bps)

        # Check statistics from breakpoints etc found by Muggeo
        # 1. MUggeo rss for these brreakpoints are
        muggeo_rss = 190.02

        self.assertAlmostEqual(
            muggeo_rss, next_fit.residual_sum_squares, places=1)

        muggeo_c = 100.64655
        muggeo_c_se = 0.23081
        muggeo_c_t = 436.07

        muggeo_alpha = -4.18526
        muggeo_alpha_se = 0.05583
        muggeo_alpha_t = -74.97

        muggeo_beta1 = -3.65462
        muggeo_beta1_se = 0.11405
        muggeo_beta1_t = -32.05

        muggeo_beta2 = 3.81336
        muggeo_beta2_se = 0.11405
        muggeo_beta2_t = 34.45

        muggeo_bp1 = 7.152
        muggeo_bp1_se = 0.101

        muggeo_bp2 = 12.161
        muggeo_bp2_se = 0.095

        estimates = next_fit.estimates

        self.assertAlmostEqual(
            muggeo_c, estimates["const"]["estimate"], places=1)
        self.assertAlmostEqual(
            muggeo_alpha, estimates["alpha1"]["estimate"], places=1)
        self.assertAlmostEqual(
            muggeo_beta1, estimates["beta1"]["estimate"], places=1)
        self.assertAlmostEqual(
            muggeo_beta2, estimates["beta2"]["estimate"], places=1)
        self.assertAlmostEqual(
            muggeo_bp1, estimates["breakpoint1"]["estimate"], places=1)
        self.assertAlmostEqual(
            muggeo_bp2, estimates["breakpoint2"]["estimate"], places=1)

        self.assertAlmostEqual(muggeo_c_se, estimates["const"]["se"], places=1)
        self.assertAlmostEqual(
            muggeo_alpha_se, estimates["alpha1"]["se"], places=1)
        self.assertAlmostEqual(
            muggeo_beta1_se, estimates["beta1"]["se"], places=1)
        self.assertAlmostEqual(
            muggeo_beta2_se, estimates["beta2"]["se"], places=1)
        self.assertAlmostEqual(
            muggeo_bp1_se, estimates["breakpoint1"]["se"], places=1)
        self.assertAlmostEqual(
            muggeo_bp2_se, estimates["breakpoint2"]["se"], places=1)

        self.assertAlmostEqual(
            muggeo_c_t, estimates["const"]["t_stat"], delta=1)
        self.assertAlmostEqual(
            muggeo_alpha_t, estimates["alpha1"]["t_stat"], delta=1)
        self.assertAlmostEqual(
            muggeo_beta1_t, estimates["beta1"]["t_stat"], delta=1)
        self.assertAlmostEqual(
            muggeo_beta2_t, estimates["beta2"]["t_stat"], delta=1)

        muggeo_r_squared = 0.9991
        muggeo_adj_r_squared = 0.999

        self.assertAlmostEqual(muggeo_r_squared, next_fit.r_squared, places=2)
        self.assertAlmostEqual(muggeo_adj_r_squared,
                               next_fit.adjusted_r_squared, places=2)

    def test_against_muggeo_r_package_data_2(self):
        """
        Muggeo uses slightly different packages and methods etc, so just check
        values are very close, not exact
        Starting from Muggeo's converged breakpoint values, I am iterating
        once, slight change
        The NextBreakpoint class is very vanilla, so in this example is
        getting in some local minima
        """

        data = SourceFileLoader('data', DATA_SOURCE).load_module()

        xx = np.array(data.MUGGEO_2_XX)
        yy = np.array(data.MUGGEO_2_YY)

        # Choose some bps values from Muggeo converged values
        bps = np.array([1.608])

        next_fit = NextBreakpoints(xx, yy, bps)

        # Check statistics from breakpoints etc found by Muggeo
        # 1. MUggeo rss for these brreakpoints are
        muggeo_rss = 34.70127

        self.assertAlmostEqual(
            muggeo_rss, next_fit.residual_sum_squares, places=1)

        muggeo_c = 4.86206
        muggeo_c_se = 0.31020
        muggeo_c_t = 15.67

        muggeo_alpha = 0.94472
        muggeo_alpha_se = 0.06744
        muggeo_alpha_t = 14.01

        muggeo_beta1 = 1.95719
        muggeo_beta1_se = 0.11008
        muggeo_beta1_t = 17.78

        muggeo_bp1 = 1.608
        muggeo_bp1_se = 0.291

        estimates = next_fit.estimates

        self.assertAlmostEqual(
            muggeo_c, estimates["const"]["estimate"], places=1)
        self.assertAlmostEqual(
            muggeo_alpha, estimates["alpha1"]["estimate"], places=1)
        self.assertAlmostEqual(
            muggeo_beta1, estimates["beta1"]["estimate"], places=1)
        self.assertAlmostEqual(
            muggeo_bp1, estimates["breakpoint1"]["estimate"], places=1)

        self.assertAlmostEqual(muggeo_c_se, estimates["const"]["se"], places=1)
        self.assertAlmostEqual(
            muggeo_alpha_se, estimates["alpha1"]["se"], places=1)
        self.assertAlmostEqual(
            muggeo_beta1_se, estimates["beta1"]["se"], places=1)
        self.assertAlmostEqual(
            muggeo_bp1_se, estimates["breakpoint1"]["se"], places=1)

        self.assertAlmostEqual(
            muggeo_c_t, estimates["const"]["t_stat"], delta=1)
        self.assertAlmostEqual(
            muggeo_alpha_t, estimates["alpha1"]["t_stat"], delta=1)
        self.assertAlmostEqual(
            muggeo_beta1_t, estimates["beta1"]["t_stat"], delta=1)

        muggeo_r_squared = 0.9911
        muggeo_adj_r_squared = 0.9903

        self.assertAlmostEqual(muggeo_r_squared, next_fit.r_squared, places=2)
        self.assertAlmostEqual(muggeo_adj_r_squared,
                               next_fit.adjusted_r_squared, places=2)


if __name__ == '__main__':
    unittest.main()
